package com.aleax.gui.gesture;

import java.io.Serializable;
import java.util.*;

/**
 * Provides storage for multiple strokes and an associated action.
 *
 * @version 1.0 28 Apr 2001
 * @author Yiannis Paschalidis
 */ 
public final class Gesture implements Serializable, Cloneable
{
	/** This gesture's id. */
	public final int id;

	/** This gesture's (human-readable) description. */
	public final String description;

	private final Vector strokes = new Vector();

	public Gesture(int id, String description)
	{
		this(id, description, null);
	}

	public Gesture(int id, String description, Stroke[] strokes)
	{
		this.id=id;
		this.description=description;

		if (strokes!=null)
		{
			this.strokes.ensureCapacity(strokes.length);

			for (int i=0 ; i<strokes.length ; i++)
				this.strokes.add(strokes[i]);
		}
	}

	public Object clone()
	{
		return new Gesture(id,description,(Stroke[]) strokes.toArray(new Stroke[strokes.size()]));
	}

	/**
	 * Is there a stroke in the gesture that matches? 
	 * @return index of matching stroke (1=1st,2=2nd...) or -1 for no match
	 */
	public boolean hasStroke(Stroke s)
	{
		for (int i=0 ; i<strokes.size() ; i++)
		{
			if (s.equivalentTo((Stroke)strokes.get(i)))
				return true;
		}
		
		return false;
	}

	/**
	 * Adds a stroke to the gesture if the stroke doesn't match one that is already defined
	 * @returns true if successful, false otherwise
	 */
	public boolean add(Stroke s)
	{
		if (hasStroke(s))
		{
			return false;
		}
		
		strokes.add(s);
		return true;
	}
	
	/** Removes the stroke with the given index. */	
	public void remove(int index)
	{
		strokes.remove(index);
	}
	
	/** @return the number of defined strokes for this gesture */
	public int numStrokes()
	{
		return strokes.size();
	}
	
	/** @return the stroke at the specified index */
	public Stroke getStroke(int index)
	{
		return (Stroke)strokes.get(index);
	}
		
	/** Recalculates all of the strokes with the current settings */
	void recalculate()
	{
		for (int i=0 ; i<strokes.size() ; i++)
		{
			((Stroke)strokes.get(i)).recalculate();
		}
	}
}
