package com.aleax.gui.gesture;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.border.*;
import javax.swing.tree.*;
import java.awt.dnd.*;
import java.awt.datatransfer.*;

/**
 * UI used to define gestures when using the GestureInterface.
 *
 * <p>This panel is designed to be placed within eg. an application settings panel.
 * See the convenience method <CODE>showDialog(Component)</CODE> for a basic example on how to us it.</p>
 *
 * @version 1.0 29 Apr 2001
 * @author Yiannis Paschalidis
 */ 
public final class GestureEditPanel extends JPanel 
	implements TreeSelectionListener, DropTargetListener, DragGestureListener, DragSourceListener
{
	private final Gesture[] gestures = GestureInterface.getGestures();

	private final JCheckBox trainingMode = new JCheckBox("Training mode active", GestureInterface.trainingMode);
	private final JCheckBox enabled = new JCheckBox("Gesture recognition enabled", GestureInterface.isEnabled());
	private final StrokePanel strokePanel = new StrokePanel(250, 250);

	private final GridBagLayout gbl = new GridBagLayout();
	private final GridBagConstraints gbc = new GridBagConstraints();

	private final XTree tree = new XTree(new XRootNode("Gestures"));

	public GestureEditPanel(Stroke newStroke)
	{
		setLayout(gbl);

		setBorder(new EmptyBorder(10, 10, 10, 10));

		//Set up the rest of the panel...
		JScrollPane scrollPane = new JScrollPane(tree);
		scrollPane.setPreferredSize(strokePanel.getPreferredSize());
		add(new JLabel("Available gestures"), 0, 0, 1, 1, GridBagConstraints.NONE, GridBagConstraints.CENTER);
		add(scrollPane, 0, 1, 1, 1, GridBagConstraints.BOTH, GridBagConstraints.WEST);

		add(new JLabel(" "), 1, 1, 1, 1, GridBagConstraints.BOTH, GridBagConstraints.WEST);

		add(new JLabel("Currently selected stroke"), 2, 0, 1, 1, GridBagConstraints.NONE, GridBagConstraints.CENTER);
		add(strokePanel, 2, 1, 1, 1, GridBagConstraints.BOTH, GridBagConstraints.WEST);
		strokePanel.setBorder(new LineBorder(Color.black));

		add(enabled, 0, 2, 1, 1, GridBagConstraints.NONE, GridBagConstraints.WEST);
		add(trainingMode, 0, 3, 1, 1, GridBagConstraints.NONE, GridBagConstraints.WEST);

		//Set up the tree...
		XRootNode root = (XRootNode)tree.getRootNode();

		if (newStroke!=null)
			root.add(new StrokeNode("New stroke", newStroke));

		for (int i=0 ; i<gestures.length ; i++)
		{
			root.add(new GestureNode(gestures[i]));
		}
		
		tree.addTreeSelectionListener(this);
		root.setExpanded(true);

		if (newStroke!=null)
		{
			((XNode)root.getChildAt(0)).setSelected(true);
		}
		
		//Init DND stuff
		DropTarget dropTarget = new DropTarget(tree, this);

		DragSource dragSource = new DragSource();
		dragSource.createDefaultDragGestureRecognizer(tree, DnDConstants.ACTION_MOVE, this);
	}

	/** 
	 * A convenience method to display a dialog with the settings panel.
	 * A modal dialog is shown with ok/cancel buttons and settings take effect
	 * immediately if the user chooses 'ok'.
	 */

	public static void showDialog(Component invoker, Stroke newStroke)
	{
		GestureEditPanel settings = new GestureEditPanel(newStroke);

		int id = JOptionPane.showOptionDialog(invoker, settings, "Gesture settings", -1,
			JOptionPane.PLAIN_MESSAGE, null, new String[]{"OK", "Cancel"}, null);

		if (id == 0) 
		{
			//Ok button
			settings.applySettings();
		}
	}

	/** Applies the changes made in the panel. */
	public void applySettings()
	{
		GestureInterface.setGestures(gestures);
		GestureInterface.trainingMode = trainingMode.isSelected();
		GestureInterface.setEnabled(enabled.isSelected());
	}

	/** Adds the given component to the panel using the specified constraints. */
	private void add(Component c, int gridx, int gridy, int gridwidth, int gridheight, int fill, int anchor)
	{
		gbc.weightx = (fill==GridBagConstraints.HORIZONTAL || fill==GridBagConstraints.BOTH) ? 1.0 : 0.0;
		gbc.weighty = (fill==GridBagConstraints.VERTICAL || fill==GridBagConstraints.BOTH) ? 1.0 : 0.0;

		gbc.gridx=gridx;
		gbc.gridy=gridy;
		gbc.gridwidth=gridwidth;
		gbc.gridheight=gridheight;
		gbc.fill=fill;
		gbc.anchor=anchor;

      gbl.setConstraints(c, gbc);
      super.add(c);
	}

	public void valueChanged(TreeSelectionEvent tse)
	{
		TreePath path=tse.getNewLeadSelectionPath();
		Stroke selectedStroke = null;

		if (path!=null)
		{
			Object node = path.getLastPathComponent();

			if (node instanceof StrokeNode)
			{
				selectedStroke = ((StrokeNode)node).stroke;
			}
		}

		strokePanel.setStroke(selectedStroke);
	}

	//----------------------------------------------------------------------------------------------
	//DND stuff

	public void dragGestureRecognized(DragGestureEvent dge) 
	{	
		Object node = tree.getSelectedNode();

		if (node instanceof StrokeNode)
		{
			// as the name suggests, starts the dragging
			dge.getDragSource().startDrag(dge, DragSource.DefaultMoveDrop, new ObjectTransferable(node), this);
		}
	}

	public void dragExit(DropTargetEvent e) {}
	public void dropActionChanged(DropTargetDragEvent e) {}

	public void dragEnter(DropTargetDragEvent e) 
	{		
		if (tree == e.getDropTargetContext().getComponent())
		{
			DataFlavor[] flavors = e.getCurrentDataFlavors();

			for (int i=0 ; i<flavors.length ; i++)
			{
				if (flavors[i].getRepresentationClass()==StrokeNode.class)
				{
					e.acceptDrag(DnDConstants.ACTION_MOVE);
					return;
				}
			}
		}
		
		e.rejectDrag();
	}
		
	public void dragOver(DropTargetDragEvent e) 
	{
		Component c = e.getDropTargetContext().getComponent();

		if (c==tree)
		{
			//Move the selection in the tree 
			//to show where the node will end up

			Point p = e.getLocation();
			TreePath path = tree.getPathForLocation(p.x, p.y);

			if (path!=null)
			{
				XNode node = (XNode)path.getLastPathComponent();

				if (node instanceof GestureNode)
					node.setSelected(true);
				else if (node.getParent() instanceof GestureNode)
					((XNode)node.getParent()).setSelected(true);
			}
		}
	}

	public void drop(DropTargetDropEvent e) 
	{
		try 
		{			
			Transferable tr = e.getTransferable();
			Component c = e.getDropTargetContext().getComponent();
			
			// Determine the flavor supported by the data source so we
			// know how to retrieve the value from the transferable object.

			if (c==tree && tree.getSelectedNode() instanceof GestureNode)
			{
				DataFlavor[] flavors = e.getCurrentDataFlavors();

				for (int i=0 ; i<flavors.length ; i++)
				{
					if (flavors[i].getRepresentationClass()==StrokeNode.class)
					{
						e.acceptDrop(DnDConstants.ACTION_MOVE);
						StrokeNode strokeNode = (StrokeNode)tr.getTransferData(flavors[i]);

						//First remove the stroke from its current gesture (if any)
						if (strokeNode.getParent() instanceof GestureNode)
						{
							GestureNode gestureNode = (GestureNode)strokeNode.getParent();
							gestureNode.gesture.remove(gestureNode.getIndex(strokeNode));
						}

						//Remove it from the tree
						strokeNode.removeFromParent();

						//Now add it to its new gesture
						GestureNode gestureNode = (GestureNode)tree.getSelectedNode();
						strokeNode.setUserObject("*Stroke #"+(gestureNode.getChildCount()+1));
						gestureNode.add(strokeNode);
						gestureNode.gesture.add(strokeNode.stroke);

						//Update the tree
						tree.forceTreeUpdate(tree.getRootNode());
						gestureNode.setExpanded(true);

						//Finally, indicate that the drop has been completed
						e.getDropTargetContext().dropComplete(true);
						return;
					}
				}
			}

			e.rejectDrop();
		}
		catch (Exception ex) 
		{
			// If an exception is raised reject the drop
			ex.printStackTrace();
			e.rejectDrop();
		}
	}

	public void dragDropEnd(DragSourceDropEvent e){}
	public void dragEnter(DragSourceDragEvent e){}
	public void dragExit(DragSourceEvent e){}
	public void dragOver(DragSourceDragEvent e){}
	public void dropActionChanged(DragSourceDragEvent e){}

	private static final class ObjectTransferable implements Transferable
	{
		private final DataFlavor flavor;
		private final Object object;
		
		ObjectTransferable(Object object)
		{
			flavor = new DataFlavor(object.getClass(), object.getClass().getName())
			{
				public String getMimeType() { return "application/"+DataFlavor.javaJVMLocalObjectMimeType; }
				public boolean isFlavorJavaFileListType() { return false; }
				public boolean isFlavorRemoteObjectType() { return false; }
				public boolean isFlavorSerializedObjectType() { return false; }
			};

			this.object = object;
		}
		
		public DataFlavor[] getTransferDataFlavors() 
		{
			return new DataFlavor[]{flavor};
		}
		
		public boolean isDataFlavorSupported(DataFlavor flavor) 
		{
			return flavor.getRepresentationClass() == flavor.getRepresentationClass();
		}
		
		public synchronized Object getTransferData(DataFlavor flavor) throws UnsupportedFlavorException
		{
			if (!isDataFlavorSupported(flavor))
			{
				throw new UnsupportedFlavorException(flavor);
			}
			
			return object;
		}
	}

	//-----------------------------------------------------------------------------------

	/** Simple extension of JComponent that paints a stroke on itself */

	private final class StrokePanel extends JComponent
	{
		private Stroke stroke;

		public StrokePanel(int width, int height)
		{
			setPreferredSize(new Dimension(width,height));
			setBackground(Color.white);
			setForeground(Color.black);
		}

		public void setStroke(Stroke stroke)
		{
			this.stroke=stroke;

			if (isShowing())
			{
				repaint();
			}
		}

		public void paint(Graphics g) 
		{
			Insets insets = getInsets();

			g.setColor(getBackground());
			g.fillRect(insets.left, insets.top, getWidth()-insets.left-insets.right, getHeight()-insets.top-insets.bottom);
			
			if (stroke!=null)
			{
				g.setColor(getForeground());
				stroke.paint(g);
			}

			paintBorder(g);
		}

		public boolean isOpaque() { return false; }
	}

	//----------------------------------------------------------------------------------------------
	// JTree / TreeNode extensions to provide a better api

	private static final class GestureNode extends XNode
	{
		final Gesture gesture;

		GestureNode(Gesture gesture)
		{
			super(gesture.description, true);
			this.gesture = gesture;

			for (int i=0 ; i<gesture.numStrokes() ; i++)
			{
				this.add(new StrokeNode("Stroke #"+(i+1),gesture.getStroke(i)));
			}
		}
	}

	private static final class StrokeNode extends XNode
	{
		final Stroke stroke;

		StrokeNode(String description, Stroke stroke)
		{
			super(description,false);
			this.stroke=stroke;
		}
	}

	private static class XNode extends DefaultMutableTreeNode
	{
		protected boolean expanded;

		XNode(String title, boolean allowsChildren)
		{
			super(title, allowsChildren);
			expanded=false;
		}

		boolean isExpanded() { return expanded; }

		void setExpanded(final boolean expanded)
		{
			this.expanded=expanded;
			final TreePath path = new TreePath(getPath());

			TreeNode root = getRoot();

			if (!(root instanceof XRootNode))
				return;

			final XTree tree = ((XRootNode) root).getTree();

			if (tree == null) 
			{
				return;
			}
			
			SwingUtilities.invokeLater
			(
				new Runnable()
				{
					public void run() 
					{
						if (expanded)
						{
							tree.expandPath(path);
						}
						else
						{
							tree.collapsePath(path);
						}
					}
				} 
			);
		}

		void setSelected(final boolean selected)
		{
			final XTree tree = ((XRootNode) getRoot()).getTree();
			final TreePath path = new TreePath(XNode.this.getPath());

			if (tree == null) 
			{
				return;
			}
			
			SwingUtilities.invokeLater
			(
				new Runnable()
				{
					public void run() 
					{
						if (selected)
						{
							tree.setSelectionPath(path);
						}
						else
						{
							tree.removeSelectionPath(path);
						}
					}
				} 
			);
		}

		void refreshTree()
		{
			TreeNode root = getRoot();

			if (!(root instanceof XRootNode))
			{
				return;
			}
			
			((XRootNode) root).getTree().forceTreeUpdate(this);
		}
	}

	private static final class XRootNode extends XNode
	{
		private XTree tree;

		XRootNode(String title)
		{
			super(title, true);
		}

		void setTree(XTree tree)	{ this.tree=tree;	}
		XTree getTree()				{ return tree; }
	}

	private static final class XTree extends JTree implements TreeExpansionListener
	{
		XTree(XRootNode root)
		{
			super(root);
			root.setTree(this);

			addTreeExpansionListener(this);

			ToolTipManager.sharedInstance().registerComponent(this);
			putClientProperty( "JTree.lineStyle", "Angled" );
			getSelectionModel().setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);
		}

		TreeNode getRootNode()
		{
			return (TreeNode) super.getModel().getRoot();
		}

		void forceTreeUpdate(final TreeNode node)
		{
			SwingUtilities.invokeLater
			(
				new Runnable()
				{
					public void run() { ((DefaultTreeModel) getModel()).reload(node); }
				}
			);
		}

		/** @return the selected Node, or null if none. */
		Object getSelectedNode()
		{
			TreePath treePath = super.getSelectionPath();

			if (treePath == null) 
				return null;

			return treePath.getLastPathComponent();
		}

		public void treeExpanded(TreeExpansionEvent event)
		{
			TreeNode node = (TreeNode) event.getPath().getLastPathComponent();
			
			if (node instanceof XNode)
			{
				((XNode) node).setExpanded(true);
			}
		}

		public void treeCollapsed(TreeExpansionEvent event)
		{
			TreeNode node = (TreeNode) event.getPath().getLastPathComponent();
			
			if (node instanceof XNode)
			{
				((XNode) node).setExpanded(false);
			}
		}
	}
}