package com.aleax.gui.gesture;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.util.zip.*;
import java.io.*;

/**
 * GestureInterface: Provides easy access to gesturing
 *
 * @version 1.0 28 Apr 2001
 * @author Yiannis Paschalidis
 */ 
public final class GestureInterface
{
	private static final GestureInterface instance = new GestureInterface();
	private final MListener listener = new MListener();;

	/** Contains all the defined gestures */
	private static final Vector gestures = new Vector();

	/** Contains all the listeners to notify when a gesture is made */
	private static final Vector listeners = new Vector();

	/** Contains all the components we're listening to */
	private static final Vector listeningTo = new Vector();

	/** Whether the training mode is active or not */
	public static boolean trainingMode = true;

	/** Whether the gesturing system is enabled */
	private static boolean enabled = true;

	/** Whether a stroke is currently being drawn */
	private static boolean drawingStroke = false;
	
	/** Keeps track of mouse movement for a particular drag */
	private static final Vector readPoints = new Vector();

	/** Hide the constructor as there's no public instance methods here */
	private GestureInterface(){}

	//----------------------------------------------------------------------------------------------
	//Methods for persistance

	/** Reads the gesture system settings from the given inputStream */

	public static void readSettings(InputStream is) throws Exception
	{
		GZIPInputStream gzis = null;
		ObjectInputStream ois = null;

		try
		{
			gzis = new GZIPInputStream(is);
			ois = new ObjectInputStream(gzis);

			//First read the system settings
			enabled = ((Boolean)ois.readObject()).booleanValue();
			trainingMode = ((Boolean)ois.readObject()).booleanValue();

			//Now read the stroke settings...		
			Stroke.compareTolerance = ((Double)ois.readObject()).doubleValue();
			Stroke.uniquenessRadius = ((Integer)ois.readObject()).intValue();
			Stroke.filterTolerance = ((Integer)ois.readObject()).intValue();
			Stroke.filterMethod = ((Integer)ois.readObject()).intValue();

			//Now read in the gestures
			gestures.clear();
			gestures.addAll((Vector)ois.readObject());
			recalculateAll();

			ois.close();
			gzis.close();
		}
		catch (Exception e)
		{
			if (ois!=null)
			{
				try { ois.close(); }
				catch (Exception ignored) {}
			}

			if (gzis!=null)
			{
				try { gzis.close(); }
				catch (Exception ignored) {}
			}

			throw e;
		}
	}

	/** Writes the gesture system settings to the given outputStream */
	public static void writeSettings(OutputStream out) throws Exception
	{
		GZIPOutputStream gzos = new GZIPOutputStream(out);
		ObjectOutputStream s = new ObjectOutputStream(gzos);

		//First write out the system settings...
		s.writeObject(Boolean.valueOf(enabled));
		s.writeObject(Boolean.valueOf(trainingMode));
      
		//Now write out the stroke settings...		
		s.writeObject(new Double(Stroke.compareTolerance));
		s.writeObject(new Integer(Stroke.uniquenessRadius));
		s.writeObject(new Integer(Stroke.filterTolerance));
		s.writeObject(new Integer(Stroke.filterMethod));
		
		//Now write out the gestures...
		s.writeObject(gestures);

		s.flush();
		s.close();
		gzos.close();
	}

	//----------------------------------------------------------------------------------------------
	//Methods for gesture definition/editing

	/** @return a copy of the gestures for use with external editing of gestures */

	public static Gesture[] getGestures()
	{
		Gesture[] clones = new Gesture[gestures.size()];

		for (int i=0 ; i<gestures.size() ; i++)
		{
			clones[i] = (Gesture) ((Gesture)gestures.get(i)).clone();
		}
		
		return clones;
	}

	public static void setGestures(Gesture[] newGestures)
	{
		gestures.clear();

		for (int i=0 ; i<newGestures.length ; i++)
		{
			gestures.add(newGestures[i].clone());
		}
	}

	/** Recalculate all the strokes when one or more filter settings have changed */

	public static void recalculateAll()
	{
		for (int i=0 ; i<gestures.size() ; i++)
		{
			((Gesture)gestures.get(i)).recalculate();
		}
	}

	//----------------------------------------------------------------------------------------------
	//Gesture enabling / disabling

	public static void addGestureListener(GestureListener listener)
	{
		if (!listeners.contains(listener))
		{
			listeners.add(listener);
		}
	}

	public static void removeGestureListener(GestureListener listener)
	{
		listeners.remove(listener);
	}

	public static void enableGestures(Component c)
	{
		//First check that we're not already listening to this component.
		//Why? Because if you add a listener multiple times to a component, it 
		//gets the events multiple times, and it can also lead to memory leaks.

		if (!listeningTo.contains(c))
		{
			c.addMouseListener(instance.listener);
			c.addMouseMotionListener(instance.listener);
			listeningTo.add(c);
		}
	}

	public static void disableGestures(Component c)
	{
		c.removeMouseListener(instance.listener);
		c.removeMouseMotionListener(instance.listener);
		listeningTo.remove(c);
	}

	public static void setEnabled(boolean enabled)
	{
		GestureInterface.enabled=enabled;

		if (enabled && drawingStroke)
		{
			cancelStroke();
		}
	}

	public static boolean isEnabled()
	{
		return enabled;
	}

	//----------------------------------------------------------------------------------------------
	//Gesture recognition

	public static void cancelStroke()
	{
		drawingStroke=false;
		readPoints.clear();
	}

	public static void beginStroke()
	{
		if (enabled)
		{
			if (drawingStroke)
			{
				throw new IllegalStateException("Attempt made to begin a stroke when one is already in progress");
			}
			
			drawingStroke=true;
			readPoints.clear();
		}
	}

	public static void continueStroke(Point p)
	{
		if (enabled)
		{
			if (!drawingStroke)
			{
				throw new IllegalStateException("Attempt made to continue a non-existant stroke");
			}
			
			if (readPoints.isEmpty() || !p.equals(readPoints.lastElement()))
			{
				readPoints.add(p);
			}
		}
	}

	public static void completeStroke()
	{
		if (enabled)
		{
			if (!drawingStroke)
				throw new IllegalStateException("Attempt made to complete a non-existant stroke");

			drawingStroke = false;

			//Was it not really a stroke?
			if (readPoints.size()<2)
				return;

			Stroke stroke = new Stroke(readPoints);
			Gesture gesture = recogniseStroke(stroke);

			if (gesture==null)
			{
				if (trainingMode)
				{
					GestureEditPanel.showDialog(null,stroke);
				}
			}
			else
			{
				for (int i=0 ; i<listeners.size() ; i++)
				{
					((GestureListener)listeners.get(i)).gesturePerformed(stroke, gesture.id);
				}
			}
		}
	}
	
	public static Gesture recogniseStroke(Stroke stroke)
	{
		for (int i=0 ; i<gestures.size() ; i++)
		{
			Gesture gesture = (Gesture) gestures.get(i);

			if (gesture.hasStroke(stroke))
			{
				return gesture;
			}
		}
		
		//nope, didn't find it anywhere
		return null;
	}

	//----------------------------------------------------------------------------------------------
	// The mouse listening happens in a private inner class to avoid being publicly accessible

	private final class MListener implements MouseListener, MouseMotionListener
	{
		//MouseListener/MouseMotionListener implementation

		public void mouseClicked(MouseEvent me) {}
		public void mouseEntered(MouseEvent me) {}
		public void mouseExited(MouseEvent me) {}
		public void mouseMoved(MouseEvent me) {}

		public void mousePressed(MouseEvent me) 
		{
			beginStroke();
			//continueStroke(screenLocation(me));
			continueStroke(me.getPoint());
		}

		public void mouseDragged(MouseEvent me) 
		{
			//continueStroke(screenLocation(me));
			continueStroke(me.getPoint());
		}
		
		public void mouseReleased(MouseEvent me) 
		{
			//continueStroke(screenLocation(me));
			continueStroke(me.getPoint());
			completeStroke();
		}

		/** @return the screen coordinate where the given mouse event happened */

		private Point screenLocation(MouseEvent me)
		{
			Point p = new Point(me.getPoint());
			Point p2 = me.getComponent().getLocationOnScreen();

			p.translate(p2.x,p2.y);
			return p;
		}
	}
}
