package com.aleax.gui.gesture;

import java.awt.*;
import javax.swing.*;
import javax.swing.border.*;
import javax.swing.text.*;

/**
 * UI to tweak recognition settings.
 *
 * <p>This panel is designed to be placed within eg. an application settings panel.
 * It doesn't rely on the GestureInterface implementation, so can be used when 
 * your application handles stroke storage/recognition itself.</p>
 *
 * <p>See the convenience method <CODE>showDialog(Component)</CODE> for a basic example on how to us it.</p>
 *
 * @version 1.0 29 Apr 2001
 * @author Yiannis Paschalidis
 */ 
public final class GestureSettingsPanel extends JPanel
{
	private final IntegerTextField uniquenessRadius = new IntegerTextField(0, 999);
	private final IntegerTextField filterTolerance = new IntegerTextField(0, 999);
	private final JSlider compareTolerance = new JSlider(0, 100, (int)(Stroke.compareTolerance*100));
	private final GridBagLayout gbl = new GridBagLayout();
	private final GridBagConstraints gbc = new GridBagConstraints();

	public GestureSettingsPanel()
	{
		setLayout(gbl);
		setBorder(new EmptyBorder(10, 10, 10, 10));

		uniquenessRadius.setText(String.valueOf(Stroke.uniquenessRadius));
		filterTolerance.setText(String.valueOf(Stroke.filterTolerance));

		add(new JLabel("Minimum segment length:  "), 0, 0, 1, 1, GridBagConstraints.NONE, GridBagConstraints.WEST);
		add(uniquenessRadius, 1, 0, 1, 1, GridBagConstraints.NONE, GridBagConstraints.WEST);
		add(new JLabel(" pixels     "), 2, 0, 1, 1, GridBagConstraints.NONE, GridBagConstraints.WEST);

		add(new JLabel("Filter tolerance:"), 0, 1, 1, 1, GridBagConstraints.NONE, GridBagConstraints.WEST);
		add(filterTolerance, 1, 1, 1, 1, GridBagConstraints.NONE, GridBagConstraints.WEST);
		add(new JLabel(" pixels"), 2, 1, 1, 1, GridBagConstraints.NONE, GridBagConstraints.WEST);

		JPanel sliderPanel = new JPanel(new BorderLayout());
		sliderPanel.add(new JLabel("Recognition tolerance", SwingConstants.CENTER), BorderLayout.NORTH);
		sliderPanel.add(new JLabel("Low"), BorderLayout.WEST);
		sliderPanel.add(compareTolerance, BorderLayout.CENTER);
		sliderPanel.add(new JLabel("High"), BorderLayout.EAST);
		sliderPanel.setBorder(new EmptyBorder(20, 0, 0, 0));

		add(sliderPanel, 0, 2, 3, 1, GridBagConstraints.NONE, GridBagConstraints.CENTER);
	}

	/** 
	 * A convenience method to display a dialog with the settings panel.
	 * A modal dialog is shown with ok/cancel buttons and settings take effect
	 * immediately if the user chooses 'ok'.
	 */

	public static void showDialog(Component invoker)
	{
		GestureSettingsPanel settings = new GestureSettingsPanel();

		int id = JOptionPane.showOptionDialog(invoker, settings, "Gesture settings", -1, 
			JOptionPane.PLAIN_MESSAGE, null, new String[]{"OK", "Cancel"}, null);

		if (id == 0) //Ok button
		{
			settings.applySettings();
		}
	}

	/** Applies the changes made in the panel. */
	public void applySettings()
	{
		Stroke.uniquenessRadius = Integer.parseInt(uniquenessRadius.getText());
		Stroke.filterTolerance = Integer.parseInt(filterTolerance.getText());
		Stroke.compareTolerance = (double)compareTolerance.getValue()/100.0;
		GestureInterface.recalculateAll();
	}

	/** Adds the given component to the panel using the specified constraints. */
	private void add(Component c, int gridx, int gridy, int gridwidth, int gridheight, int fill, int anchor)
	{
		gbc.weightx = (fill==GridBagConstraints.HORIZONTAL || fill==GridBagConstraints.BOTH) ? 1.0 : 0.0;
		gbc.weighty = (fill==GridBagConstraints.VERTICAL || fill==GridBagConstraints.BOTH) ? 1.0 : 0.0;

		gbc.gridx = gridx;
		gbc.gridy = gridy;
		gbc.gridwidth = gridwidth;
		gbc.gridheight = gridheight;
		gbc.fill = fill;
		gbc.anchor = anchor;

      gbl.setConstraints(c, gbc);
      super.add(c);
	}

	//-----------------------------------------------------------------------------------
	
	/** Extends JTextField to only accept integer values */

	private static final class IntegerTextField extends JTextField
	{
		IntegerTextField()
		{
			this(Integer.MIN_VALUE, Integer.MAX_VALUE);
		}

		IntegerTextField(int minValue, int maxValue)
		{
			super
			(
				Math.max(String.valueOf(minValue).length(), 
							String.valueOf(maxValue).length())
			);

			if (minValue > maxValue)
			{
				throw new IllegalArgumentException("minValue>maxValue");
			}
			
			setDocument(new IntegerDocument(minValue, maxValue));
			setText(String.valueOf(minValue));
		}

		//--------------------------------------------------------------------------------
		// PENDING: This doesn't handle deletes and the minimum value very well

		private static final class IntegerDocument extends PlainDocument 
		{		
			private final int minValue;
			private final int maxValue;

			IntegerDocument(int minValue, int maxValue)
			{
				this.minValue = minValue;
				this.maxValue = maxValue;
			}

			public void insertString(int offs, String str, AttributeSet a)
				throws BadLocationException 
			{	
				String text = getText(0, getLength());
				text = text.substring(0, offs) + str + text.substring(offs);

				try
				{
					int newValue = Integer.parseInt(text);

					if (newValue >= minValue && newValue <= maxValue)
					{
						super.insertString(offs, str, a);
						return;
					}
				}
				catch (Exception ignored){}

				Toolkit.getDefaultToolkit().beep();
			}
		}	
	}
}
