package com.aleax.gui.gesture;

import java.awt.*;
import java.util.*;
import java.io.Serializable;

/**
 * This class provides low-level stroke manipulation/comparison routines.
 *
 * <p>A stroke is the series of Points recorded as a drag
 * with a single, continuous movement of the pointing device.</p>
 *
 * @version 1.0 28 Apr 2001
 * @author Yiannis Paschalidis
 */ 
public final class Stroke implements Serializable, Cloneable
{
	public static final long serialVersionUID = -8100653415102101955L;

	private static final int FILTER_CUSTOM = 0;
	private static final int FILTER_ENDPOINTS = 1;

	static double compareTolerance = 0.1;
	static int uniquenessRadius = 5;
	static int filterTolerance = 20;
	static int filterMethod = FILTER_ENDPOINTS;

	//----------------------------------------------------------------------------------------------
	// Raw data

	/**
	  * The initial data read in from the mouse is kept to draw the stroke, for loading & saving,
	  * and to recalculate the stroke comparison info when required.
	  */
	private final Point[] rawPoints;

	/** The filtered data is kept to show filtering effects when the stroke is drawn */
	private transient Point[] filteredPoints = null;

	//----------------------------------------------------------------------------------------------
	// Representation of the data used for comparison.

	/** Turn angles. Will have filteredPoints.length-2 elements. */
	private transient int[] segmentAngles = null;

	/** Segment lengths relative to the longest segment (0->100%). It will have filteredPoints.length-1 elements */
	private transient int[] segmentLengths = null;

	/** 
	  * Binary representation for the first 32 turns; left turn=0, right turn=1. Is used as the hashcode
	  * and to speed up stroke comparison when there are many strokes 
	  */

	private transient int strokeId = 0;

	/** The direction of the initial segment in degrees */
	private int initialSegmentDirection = 0;

	/** The length of the initial segment in pixels */
	private int initialSegmentLength = 0;

	/** The direction of the longest segment in degrees */
	private int longestSegmentDirection = 0;

	/** The length of the longest segment in pixels */
	private int longestSegmentLength = 0;

	/** The bounds of the stroke */
	private final Rectangle bounds = new Rectangle();

	//----------------------------------------------------------------------------------------------

	/** Constructs a stroke defined by the given Vector of java.awt.Point */

	public Stroke(Vector points)
	{
		this((Point[]) points.toArray(new Point[points.size()]));
	}

	/** Constructs a stroke defined by the given points */

	public Stroke(Point[] points)
	{
		//Sanity check
		if (points.length < 2)
		{
			throw new IllegalArgumentException("Strokes must have at least a start and an end point");
		}
		
		rawPoints = points;

		//figure out the stroke's bounds
		bounds.x = rawPoints[0].x;
		bounds.y = rawPoints[0].y;

		for (int i = 0 ; i<rawPoints.length ; i++)
		{
			bounds.add(rawPoints[i]);
		}
		
		//Now filter & extract important info from the raw data
		recalculate();
	}

	public Object clone()
	{
		return new Stroke(rawPoints);
	}

	public int hashCode()
	{
		return strokeId;
	}

	public boolean equals(Object o)
	{
		//can only be equal to another stroke
		if (!(o instanceof Stroke))
		{
			return false;
		}
		
		Stroke stroke = (Stroke)o;

		//Quick check - are the stroke id's the same?
		if (strokeId != stroke.strokeId || segmentAngles.length != stroke.segmentAngles.length)
		{
			return false;
		}
		
		//Ok, check the points...
		if (rawPoints.length != stroke.rawPoints.length)
		{
			return false;
		}
		
		for (int i=0 ; i<rawPoints.length ; i++)
		{
			if (!rawPoints[i].equals(stroke.rawPoints[i]))
			{
				return false;
			}
		}

		//Ok, the strokes must be equal
		return true;
	}

	public boolean equivalentTo(Stroke other)
	{
		//First check that the stroke IDs match
		if (strokeId != other.strokeId || segmentAngles.length != other.segmentAngles.length)
		{
			return false;
		}
		
		//A straight line will always match a straight line 
		//although it seems pointless to define such a stroke
		if (segmentAngles.length == 0)
		{
			return true;
		}
		
		//Check angles first...
		//1.0 = 100% tolerant (ie. doesn't care what the angles are).
		//0.5 = 50% tolerant (average difference<=90 degrees)
		//I suggest 10% tolerance (~20 degrees)
		if (!diffsWithinTolerance(segmentAngles,other.segmentAngles,compareTolerance*180.0))
		{
			return false;
		}
		
		//Now check segment lengths...
		//1.0 = 100% tolerant (ie. doesn't care what the lengths are).
		//0.5 = 50% tolerant (average difference<=50% degrees)
		//I suggest 10% tolerance
		return diffsWithinTolerance(segmentLengths,other.segmentLengths,compareTolerance*100.0);
	}

	private boolean diffsWithinTolerance(int[] array1, int[] array2, double tolerance)
	{
		//sanity check
		if (array1.length != array2.length)
		{
			return false;
		}
		
		//Cumulative differences between items in the arrays
		int diff = 0;

		//First compare the turn angles
		for (int i = 0 ; i < array1.length ; i++)
		{
			//Square the differences so that the smaller differences don't count as much
			int diff_i = array1[i] - array2[i];
			diff += diff_i * diff_i;
		}

		double averageDiff = Math.sqrt((double)diff/(double)array1.length);

		return averageDiff <= tolerance;
	}

	/** 
	 * Returns the turn angle between 1st and 2nd lines given 3 points.
	 * It is *NOT* the actual angle between the lines (0 corresponds to a straight line).
	 * The returned value is between -180 and +180, -ve for clockwise/right direction.
	 *
	 * @param p1 the start of the first line
	 * @param p2 the intersection of the lines
	 * @param p3 the end of the second line
	 * @return the turn angle between the two lines
	 */
	private int turnAngle(Point p1, Point p2, Point p3)
	{
		int a1 = lineAngle(p1,p2);
		int a2 = lineAngle(p2,p3);
		
		//check if one or more points are coinciding
		if (a1<0 || a2<0)
		{
			return 0;
		}
		
		if (Math.abs(a2-a1) <= 180)
		{
			return a2 - a1;
		}
		else
		{
			return a1 - a2;
		}
	}

	/** 
	 * @returns the angle of the line, from 0 to 359, -ve=error.
	 * NOTE: y coordinates are negated as the screen y axis is inverted (wrt cartesian axis
	 */
	private int lineAngle(Point start, Point end)
	{
		//Special check for vertical lines...
		if (start.x == end.x)
		{
			if (start.y == end.y)
			{
				return -1; //error
			}
			else
			{
				return (start.y > end.y) ? 90 : 270;
			}
		}

		int angle = (int)(Math.atan((double)(start.y - end.y)/(double)(end.x - start.x)) * 180.0/Math.PI);

		//Correct angle for 2nd,3rd quadrants
		if (start.x > end.x)
		{
			angle += 180;
		}
		
		//Correct angle for 4th quadrant
		if (start.x < end.x && end.y > start.y)
		{
			angle += 360;
		}
		
		return angle;
	}

	/**
	 * Determines the length of a line.
	 * @param p1 the line start point.
	 * @param p2 the line end point.
	 * @returns the closest int to the line length between two points.
 	 */
	private int lineLength(Point start, Point end)
	{
		double dx = (double)(end.x - start.x);
		double dy = (double)(end.y - start.y);

		return (int)(Math.sqrt(dx*dx + dy*dy) + 0.5);
	}

	public void paint(Graphics g)
	{
		Rectangle clipRect = g.getClipBounds();

		//Set up the scaling...
		double scale = Math.min( (double)clipRect.width/(double)bounds.width,
			                      (double)clipRect.height/(double)bounds.height) * 0.8;

		//shift to the top-left and leave 10% of the viewport area as a border...
		Point offset = new Point((int)(-bounds.x*scale + clipRect.width/10), (int)(-bounds.y*scale + clipRect.height/10));

		//We want the stroke centered in the draw area...
		offset.x += (clipRect.width*0.8 - bounds.width*scale)/2;
		offset.y += (clipRect.height*0.8 - bounds.height*scale)/2;

		g.translate(offset.x,offset.y);

		//Draw the stroke...
		for (int i=0 ; i<rawPoints.length-1 ; i++)
		{
			g.drawLine((int)(rawPoints[i].x*scale),(int)(rawPoints[i].y*scale),
			           (int)(rawPoints[i+1].x*scale),(int)(rawPoints[i+1].y*scale));
		}

		//Draw the feature points...
		for (int i=0 ; i<filteredPoints.length ; i++)
		{
			g.fillRect((int)(filteredPoints[i].x*scale)-3,(int)(filteredPoints[i].y*scale)-3,7,7);
		}
		
		//Draw the start point...
		g.drawRect((int)(rawPoints[0].x*scale)-6,(int)(rawPoints[0].y*scale)-6,12,12);

		//Return the graphics context to it's previous state
		g.translate(-offset.x,-offset.y);
	}

	/** For debugging */
	public String toString()
	{
		return "Raw length: "+rawPoints.length +
				 "\nFiltered length: "+filteredPoints.length +
				 "\nStroke length: "+segmentAngles.length +
				 "\nStroke id: "+Integer.toBinaryString(strokeId);
	}

    /** Re-calculates the stroke data with the current filter settings. */
	public void recalculate()
	{
		Vector points = new Vector(rawPoints.length);
		
		int lastpoint = 0;
		
		//copy in the data
		for (int i=0 ; i<rawPoints.length-1 ; i++)
		{
			//Pre-filtering: discard points which are too close together and will give
			//significant errors when comparing to the overall line angle
			if (i == 0 || lineLength(rawPoints[lastpoint],rawPoints[i]) >= uniquenessRadius)
			{
				points.add(rawPoints[i]);
				lastpoint = i;
			}
		}
		
		//Add in the last point.
		points.add(rawPoints[rawPoints.length - 1]);
		
		//Filter the data
		filter(points);
		
		filteredPoints = (Point[]) points.toArray(new Point[points.size()]);
		
		//Generate the stroke comparison info
		generateCompareInfo();
	}

	/** Generates the comparison info */
	private void generateCompareInfo()
	{
		strokeId = 0;
		segmentAngles = new int[filteredPoints.length-2];
		segmentLengths = new int[filteredPoints.length-1];
		
		//Set the initial direction & length variables
		//if the stroke is not a single point
		if (filteredPoints.length >= 2)
		{
			initialSegmentDirection=lineAngle(filteredPoints[0],filteredPoints[1]);
			initialSegmentLength=lineLength(filteredPoints[0],filteredPoints[1]);
		}
		
		//Calculate the line segment lengths and the longest segment length
		for (int i=0 ; i<filteredPoints.length-1 ; i++)
		{
			segmentLengths[i] = lineLength(filteredPoints[i],filteredPoints[i+1]);

			if (longestSegmentLength < segmentLengths[i])
			{
				longestSegmentLength = segmentLengths[i];
				longestSegmentDirection = lineAngle(filteredPoints[i],filteredPoints[i+1]);
			}
		}
		
		//Now use the longest segment length to calculate relative segment lengths
		for (int i=0 ; i<segmentLengths.length ; i++)
		{
			segmentLengths[i]=(segmentLengths[i]*100)/longestSegmentLength;
		}
		
		//Calculate the turns. First & last points ignored as they can't have turns.
		for (int i=1 ; i<filteredPoints.length-1 ; i++)
		{
			segmentAngles[i-1] = turnAngle(filteredPoints[i-1], filteredPoints[i], filteredPoints[i+1]);
			
			//Extract sign of the angle and use add it to the LSB of the stroke ID
			//Left (anticlockwise) turn gives 0, right turn 1
			strokeId = (strokeId<<1) | (segmentAngles[i-1]<0 ? 1 : 0);
		}
	}

	//----------------------------------------------------------------------------------------------
	// Filtering stuff...

	private void filter(Vector points)
	{
		switch (filterMethod)
		{
			case FILTER_CUSTOM:
				customFilter(points);
				break;
				
			case FILTER_ENDPOINTS:
				endPointsFilter(points,0,points.size()-1);
				break;
		}
	}

	/** Filters the stroke data using the custom filter. */                
	private void customFilter(Vector points)
	{
		int lastAngle;
		int tempAngle;
		int oldlength = points.size()+1;
		int lastpoint;
		
		//make sure we have enough data to Filter...
		if (points.size() < 2)
		{
			return;
		}
		
		while (points.size() < oldlength)
		{
			lastpoint = 0;
			oldlength = points.size();        
			
			//First remove excess points on 'straight' lines by checking whether
			//the current line segment is a 'turn' compared with the start of the line
			for (int i=1 ; i < points.size()-1 ; i++)
			{
				tempAngle=turnAngle((Point)points.get(lastpoint),
					(Point)points.get(i),
					(Point)points.get(i+1));
				
				if (Math.abs(tempAngle) <= filterTolerance)
				{
					points.remove(i--);
				}
				else
				{
					lastpoint = i;
				}
			}
		}
		
		//filter for small turns
		for (int i = 1 ; i<points.size() - 1 ; i++)
		{
			if (pointToLineDistance((Point)points.get(i-1),(Point)points.get(i+1),(Point)points.get(i))<=uniquenessRadius*1.5)
			{
				//then delete the middle point & adjust the loop counter
				points.remove(i--);
			}
		}
	}

	/** Java implementation of the iterative end-points fit algorithm (Duda and Hart 1973). */
	private void endPointsFilter(Vector points, int start, int end)
	{
		//The distance away for the furthest point
		//To speed things up a bit set it equal to the tolerance to begin with
		double maxDist = (double)filterTolerance;
		
		//which point was the furthest away
		int maxDistIndex = 0;
		
		for (int i = start + 1 ; i < end ; i++)
		{
			double temp = pointToLineDistance((Point)points.get(start),(Point)points.get(end),(Point)points.get(i));
			
			if (temp > maxDist)
			{
				maxDist=temp;
				maxDistIndex=i;
			}
		}
		
		if (maxDist > filterTolerance)
		{
			//We must recurse on the right segment first as its start & end
			//array indices might change when filtering the left part
			endPointsFilter(points,maxDistIndex,end);
			endPointsFilter(points,start,maxDistIndex);
		}
		else
		{
			//Remove all the points between the start & end points
			for (int i=1 ; i<end-start ; i++)
			{
				points.remove(start + 1);
			}
		}
	}

	/** return the (absolute) perpendicular distance from a point to a line */
	double pointToLineDistance(Point start, Point end, Point point)
	{
		//Simple case... line is a point
		if (start.equals(end))
		{
			return lineLength(start,point);
		}
		
		double dx = (double)(end.x-start.x);
		double dy = (double)(end.y-start.y);
		double s = Math.sqrt(dx*dx + dy*dy);

		s = ( (end.y-start.y) * (point.x-start.x)-
			   (end.x-start.x) * (point.y-start.y) )/s;

		return Math.abs(s);
	}
}
